import numpy as np

from sklearn.preprocessing import StandardScaler
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, Matern, RationalQuadratic, WhiteKernel, ConstantKernel as C

from fulu._base_aug import BaseAugmentation, add_log_lam


class GaussianProcessesAugmentation(BaseAugmentation):
    """
    Light Curve Augmentation based on Gaussian Processes Regression

    Parameters:
    -----------
    passband2lam : dict
        A dictionary, where key is a passband ID and value is Log10 of its wave length.
        Example:
            passband2lam  = {0: np.log10(3751.36), 1: np.log10(4741.64), 2: np.log10(6173.23),
                             3: np.log10(7501.62), 4: np.log10(8679.19), 5: np.log10(9711.53)}
    kernel : kernel object from sklearn
        Kernel for GaussianProcessRegressor. Can be combine from different kernels.
        Example:
            kernel = C(1.0)*RBF([1, 1]) + Matern() + WhiteKernel()
    use_err : bool
        Flag responsible for using flux error by GP Regressor.
    """

    def __init__(self, passband2lam, kernel=C(1.0) * RBF([1.0, 1.0]) + WhiteKernel(), use_err=False):

        super().__init__(passband2lam)

        self.ss = None
        self.reg = None
        self.kernel = kernel
        self.use_err = use_err

    def fit(self, t, flux, flux_err, passband):
        """
        Fit an augmentation model.

        Parameters:
        -----------
        t : array-like
            Timestamps of light curve observations.
        flux : array-like
            Flux of the light curve observations.
        flux_err : array-like
            Flux errors of the light curve observations.
        passband : array-like
            Passband IDs for each observation.
        """
        super().fit(t, flux, flux_err, passband)

        log_lam = add_log_lam(passband, self.passband2lam)

        t = np.array(t)
        flux = np.array(flux)
        flux_err = np.array(flux_err)
        passband = np.array(passband)
        log_lam = add_log_lam(passband, self.passband2lam)

        X = np.concatenate((t.reshape(-1, 1), log_lam.reshape(-1, 1)), axis=1)

        X_error = (flux_err / np.std(flux)).reshape(-1)
        self.ss = StandardScaler()
        X_ss = self.ss.fit_transform(X)

        reg_kwargs = dict(
            kernel=self.kernel, optimizer="fmin_l_bfgs_b", n_restarts_optimizer=5, normalize_y=True, random_state=42
        )

        if self.use_err:
            reg_kwargs["alpha"] = X_error**2

        self.reg = GaussianProcessRegressor(**reg_kwargs)

        self.reg.fit(X_ss, flux)
        return self

    def predict(self, t, passband):
        """
        Apply the augmentation model to the given observation time moments.

        Parameters:
        -----------
        t : array-like
            Timestamps of light curve observations.
        passband : array-like
            Passband IDs for each observation.

        Returns:
        --------
        flux_pred : array-like
            Flux of the light curve observations, approximated by the augmentation model.
        flux_err_pred : array-like
            Flux errors of the light curve observations, estimated by the augmentation model.
        """

        t = np.array(t)
        passband = np.array(passband)
        log_lam = add_log_lam(passband, self.passband2lam)

        X = np.concatenate((t.reshape(-1, 1), log_lam.reshape(-1, 1)), axis=1)
        X_ss = self.ss.transform(X)

        flux_pred, flux_err_pred = self.reg.predict(X_ss, return_std=True)

        return flux_pred, flux_err_pred



import numpy as np

import torch
import torch.nn as nn
from torch.utils.data import TensorDataset, DataLoader
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import mean_squared_error

from fulu._base_aug import BaseAugmentation, add_log_lam


class NNRegressor(nn.Module):
    def __init__(self, n_inputs=1, n_hidden=10, activation="tanh"):
        super(NNRegressor, self).__init__()

        act = nn.Tanh()
        if activation == "tanh":
            act = nn.Tanh()
        elif activation == "relu":
            act = nn.ReLU()
        elif activation == "sigmoid":
            act = nn.Sigmoid()
        else:
            raise ValueError("activation function {} is not supported".format(activation))

        self.seq = nn.Sequential(nn.Linear(n_inputs, n_hidden), act, nn.Linear(n_hidden, 1))

    def forward(self, x):
        return self.seq(x)


class FitNNRegressor:
    def __init__(
        self,
        n_hidden=10,
        activation="tanh",
        n_epochs=10,
        batch_size=64,
        lr=0.01,
        weight_decay=0.0,
        optimizer="Adam",
        debug=0,
        device="auto",
    ):
        self.model = None
        self.n_hidden = n_hidden
        self.activation = activation
        self.n_epochs = n_epochs
        self.batch_size = batch_size
        self.lr = lr
        self.weight_decay = weight_decay
        self.optimizer = optimizer
        self.debug = debug

        if device == "auto":
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = torch.device(device)

    def fit(self, X, y):
        # Estimate model
        self.model = NNRegressor(n_inputs=X.shape[1], n_hidden=self.n_hidden, activation=self.activation).to(
            self.device
        )
        # Convert X and y into torch tensors
        X_tensor = torch.as_tensor(X, dtype=torch.float32, device=self.device)
        y_tensor = torch.as_tensor(y, dtype=torch.float32, device=self.device)
        # Create dataset for trainig procedure
        train_data = TensorDataset(X_tensor, y_tensor)
        # Estimate loss
        loss_func = nn.MSELoss()
        # Estimate optimizer
        if self.optimizer == "Adam":
            opt = torch.optim.Adam(self.model.parameters(), lr=self.lr, weight_decay=self.weight_decay)
        elif self.optimizer == "SGD":
            opt = torch.optim.SGD(self.model.parameters(), lr=self.lr, weight_decay=self.weight_decay)
        elif self.optimizer == "RMSprop":
            opt = torch.optim.RMSprop(self.model.parameters(), lr=self.lr, weight_decay=self.weight_decay)
        else:
            raise ValueError('optimizer "{}" is not supported'.format(self.optimizer))
        # Enable dropout
        self.model.train(True)
        # Start the model fit
        for epoch_i in range(self.n_epochs):
            for x_batch, y_batch in DataLoader(train_data, batch_size=self.batch_size, shuffle=True):
                # make prediction on a batch
                y_pred_batch = self.model(x_batch)
                loss = loss_func(y_batch, y_pred_batch)
                # zero the parameter gradients
                opt.zero_grad()
                # backpropagate gradients
                loss.backward()
                # update the model weights
                opt.step()

    def predict(self, X):
        with torch.no_grad():
            # Disable droout
            self.model.train(False)
            # Convert X and y into torch tensors
            X_tensor = torch.as_tensor(X, dtype=torch.float32, device=self.device)
            # Make predictions for X
            y_pred = self.model(X_tensor)
            y_pred = y_pred.cpu().detach().numpy()
            return y_pred


class SingleLayerNetAugmentation(BaseAugmentation):
    """
    Light Curve Augmentation based on NNRegressor with new features

    Parameters:
    -----------
    passband2lam : dict
        A dictionary, where key is a passband ID and value is Log10 of its wave length.
        Example:
            passband2lam  = {0: np.log10(3751.36), 1: np.log10(4741.64), 2: np.log10(6173.23),
                             3: np.log10(7501.62), 4: np.log10(8679.19), 5: np.log10(9711.53)}
    n_hidden : int
        Number of neurons in a layer.
    activation : string
        Neuron's activation function. Possible values: {'tanh', 'relu', 'sigmoid'}.
    n_epochs : int
        Number of epochs of model weights optimization.
    batch_size : int
        Number of samples for one iteration of model weights optimization.
    lr : float
        Learning rate value for model weights optimization.
    optimizer : string
        Optimization algorithm. Possible values: {'SGD', 'Adam', 'RMSprop'}.
    device : str or torch device, optional
        Torch device name, default is 'auto' which uses CUDA if available and CPU if not
    weight_decay : float
        L2 penalty (regularization term) parameter.
    """

    def __init__(
        self,
        passband2lam,
        n_hidden=20,
        activation="tanh",
        n_epochs=1000,
        batch_size=500,
        lr=0.01,
        optimizer="Adam",
        device="auto",
        weight_decay=0,
    ):
        super().__init__(passband2lam)

        self.n_hidden = n_hidden
        self.activation = activation
        self.n_epochs = n_epochs
        self.batch_size = batch_size
        self.lr = lr
        self.optimizer = optimizer
        self.device = device
        self.weight_decay = weight_decay

        self.ss_x = None
        self.ss_y = None
        self.ss_t = None
        self.reg = None
        self.flux_err = 0

    def _preproc_features(self, t, passband, ss_t):
        passband = np.array(passband)
        log_lam = add_log_lam(passband, self.passband2lam)
        t = ss_t.transform(np.array(t).reshape((-1, 1)))

        X = np.concatenate((t, log_lam.reshape((-1, 1))), axis=1)
        return X

    def fit(self, t, flux, flux_err, passband):
        """
        Fit an augmentation model.

        Parameters:
        -----------
        t : array-like
            Scaled timestamps of light curve observations.
        flux : array-like
            Flux of the light curve observations.
        flux_err : array-like
            Flux errors of the light curve observations.
        passband : array-like
            Passband IDs for each observation.
        """

        super().fit(t, flux, flux_err, passband)

        self.ss_t = StandardScaler().fit(np.array(t).reshape((-1, 1)))

        X = self._preproc_features(t, passband, self.ss_t)
        self.ss_x = StandardScaler().fit(X)
        X_ss = self.ss_x.transform(X)
        flux = np.array(flux)

        self.ss_y = StandardScaler().fit(flux.reshape((-1, 1)))
        y_ss = self.ss_y.transform(flux.reshape((-1, 1)))
        self.reg = FitNNRegressor(
            n_hidden=self.n_hidden,
            activation=self.activation,
            n_epochs=self.n_epochs,
            batch_size=self.batch_size,
            lr=self.lr,
            optimizer=self.optimizer,
            device=self.device,
            weight_decay=self.weight_decay,
        )
        self.reg.fit(X_ss, y_ss)

        flux_pred = self.ss_y.inverse_transform(self.reg.predict(X_ss)).reshape(
            -1,
        )
        self.flux_err = np.sqrt(mean_squared_error(flux, flux_pred))
        return self

    def predict(self, t, passband):
        """
        Apply the augmentation model to the given observation time moments.

        Parameters:
        -----------
        t : array-like
            Scaled timestamps of light curve observations.
        passband : array-like
            Passband IDs for each observation.

        Returns:
        --------
        flux_pred : array-like
            Flux of the light curve observations, approximated by the augmentation model.
        flux_err_pred : array-like
            Flux errors of the light curve observations, estimated by the augmentation model.
        """

        X = self._preproc_features(t, passband, self.ss_t)
        X_ss = self.ss_x.transform(X)

        flux_pred = self.ss_y.inverse_transform(self.reg.predict(X_ss)).reshape(
            -1,
        )
        flux_err_pred = np.full_like(flux_pred, self.flux_err)

        return np.maximum(flux_pred, np.zeros(flux_pred.shape)), flux_err_pred



from abc import ABC, abstractmethod

import numpy as np
import matplotlib.pyplot as plt

from fulu.plotting import LcPlotter


def add_log_lam(passband, passband2lam):
    log_lam = np.array([passband2lam[i] for i in passband])
    return log_lam


def create_aug_data(t_min, t_max, passband_ids, n_obs=1000):
    t = []
    passband = []
    for band in passband_ids:
        t += list(np.linspace(t_min, t_max, n_obs))
        passband += [band] * n_obs
    return np.array(t), np.array(passband)


class BaseAugmentation(ABC):
    """
    Base abstract class for light curve augmentation

    Parameters:
    -----------
    passband2lam : dict
        A dictionary, where key is a passband ID and value is Log10 of its wave length.
        Example:
            passband2lam  = {0: np.log10(3751.36), 1: np.log10(4741.64), 2: np.log10(6173.23),
                             3: np.log10(7501.62), 4: np.log10(8679.19), 5: np.log10(9711.53)}
    """

    def __init__(self, passband2lam):
        self.passband2lam = passband2lam
        self.plotter = LcPlotter(passband2lam)
        self.t_train = None
        self.flux_train = None
        self.flux_err_train = None
        self.passband_train = None

    @abstractmethod
    def fit(self, t, flux, flux_err, passband):
        """
        Fit an augmentation model.

        This base method must be called from child classes

        Parameters:
        -----------
        t : array-like
            Timestamps of light curve observations.
        flux : array-like
            Flux of the light curve observations.
        flux_err : array-like
            Flux errors of the light curve observations.
        passband : array-like
            Passband IDs for each observation.
        """
        self.t_train = np.asarray(t)
        self.flux_train = np.asarray(flux)
        self.flux_err_train = np.asarray(flux_err)
        self.passband_train = np.asarray(passband)

    @abstractmethod
    def predict(self, t, passband):
        """
        Apply the augmentation model to the given observation time moments.

        Parameters:
        -----------
        t : array-like
            Timestamps of light curve observations.
        passband : array-like
            Passband IDs for each observation.

        Returns:
        --------
        flux_pred : array-like
            Flux of the light curve observations, approximated by the augmentation model.
        flux_err_pred : array-like
            Flux errors of the light curve observations, estimated by the augmentation model.
        """
        raise NotImplemented

    def augmentation(self, t_min, t_max, n_obs=100):
        """
        The light curve augmentation.

        Parameters:
        -----------
        t_min, t_max : float
            Min and max timestamps of light curve observations.
        n_obs : int
            Number of observations in each passband required.

        Returns:
        --------
        t_aug : array-like
            Timestamps of light curve observations.
        flux_aug : array-like
            Flux of the light curve observations, approximated by the augmentation model.
        flux_err_aug : array-like
            Flux errors of the light curve observations, estimated by the augmentation model.
        passband_aug : array-like
            Passband IDs for each observation.
        """

        t_aug, passband_aug = create_aug_data(t_min, t_max, tuple(self.passband2lam), n_obs)
        flux_aug, flux_err_aug = self.predict(t_aug, passband_aug)

        return t_aug, flux_aug, flux_err_aug, passband_aug

    def _plot_passband(self, passband, ax, approx=None):
        """
        Helper to construct a light curve in the next method plotter.

        Parameters:
        -----------
        passband : str or int or float
            A key of self.passband2lam dict.
        ax : matplotlib.pyplot.subplot object
            You can set the axis as an element of your matplotlib.pyplot.figure object.
        approx : tuple of array-like or None
            Augumentated light curve.
        """

        self.plotter.errorbar_passband(
            t=self.t_train,
            flux=self.flux_train,
            flux_err=self.flux_err_train,
            passbands=self.passband_train,
            passband=passband,
            ax=ax,
        )

        if approx:
            anobject_approx = self.plotter._make_dataframe(*approx)
            anobject_approx = anobject_approx.sort_values("time")
            light_curve_approx = anobject_approx[anobject_approx.passband == passband]
            ax.plot(
                light_curve_approx["time"].values,
                light_curve_approx["flux"].values,
                linewidth=3.5,
                color=self.plotter.colors[passband],
                label="{} approx flux".format(passband),
                zorder=10,
            )
            ax.fill_between(
                light_curve_approx["time"].values,
                light_curve_approx["flux"].values - light_curve_approx["flux_err"].values,
                light_curve_approx["flux"].values + light_curve_approx["flux_err"].values,
                color=self.plotter.colors[passband],
                alpha=0.2,
                label="{} approx sigma".format(passband),
            )

    def plot(
        self,
        *,
        plot_approx=True,
        n_approx=1000,
        passband=None,
        ax=None,
        true_peak=None,
        plot_peak=False,
        title="",
        save=None,
    ):
        """
        Plotting train points of light curve with errors for all passbands on one graph by default. A black solid curve isn't plotted at the predicted points. The predicted flux errors are also plotted using a gray bar.

        If you submit the name passband, only the submitted passband will built.

        Parameters:
        -----------
        plot_approx : bool, optional
            Flag indicating it is required to plot an approximation curve or
            isn't
        n_approx : int
            Number of approximation points in each passband.
        passband : optional
            Key of self.passband2lam dict or `None`, plot all passbands in the
            latter case
        ax : matplotlib.pyplot.subplot or None, optional
            `matplotlib` axis to plot the graph on, if `None` then new figure
            will be created and corresponded axis object will be returned
        true_peak : float or None, optional
            Plot the real peak time moment of the light curve
        plot_peak : bool, optional
            Plot approximated model peak
        title : str, optional
            The name of the graph set by `ax`.
        save : str or None, optional
            The path for saving graph.

        Returns
        -------
        ax : matplotlib.pyplot.subplot
            `matplotlib` axis with the graph
        """

        if ax is None:
            ax = self.plotter._ax_adjust()

        approx = self.augmentation(np.min(self.t_train), np.max(self.t_train), n_approx)
        if plot_approx:
            plot_approx = approx

        if passband is None:
            for band in self.passband2lam:
                self._plot_passband(band, ax, plot_approx)
        else:
            self._plot_passband(passband, ax, plot_approx)

        if true_peak is not None:
            self.plotter.plot_true_peak(true_peak=true_peak, ax=ax)

        if plot_peak:
            self.plotter.plot_sum_passbands(*approx, ax=ax)
            self.plotter.plot_peak(*approx, ax=ax)

        ax.set_title(title, size=35, pad=15)
        ax.legend(loc="best", ncol=3, fontsize=20)
        if save is not None:
            self.plotter._save_fig(save)

        return ax



import numpy as np
import matplotlib
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import pandas as pd


class LcPlotter:
    def __init__(self, passband2lam):
        self.passband2lam = passband2lam
        self.colors = {key: tone for key, tone in zip(passband2lam.keys(), mcolors.TABLEAU_COLORS.keys())}

    @staticmethod
    def _make_dataframe(t, flux, flux_err, passband):
        if t is None or flux is None or flux_err is None or passband is None:
            raise ValueError("All values must be numerical arrays")

        obj = pd.DataFrame()
        obj["time"] = t
        obj["flux"] = flux
        obj["flux_err"] = flux_err
        obj["passband"] = passband
        return obj

    @staticmethod
    def _ax_adjust():
        fig = plt.figure(figsize=(20, 10), dpi=100)
        plt.rcParams.update({"font.size": 30})
        fig.subplots_adjust(left=0.07, right=0.98, top=0.90, bottom=0.15)
        ax = plt.subplot(1, 1, 1)
        ax.spines["bottom"].set_linewidth(3)
        ax.spines["top"].set_linewidth(3)
        ax.spines["left"].set_linewidth(3)
        ax.spines["right"].set_linewidth(3)
        ax.set_xlabel("Time", size=30)
        ax.set_ylabel("Flux", size=30)
        ax.grid(linewidth=2)

        return ax

    @staticmethod
    def _save_fig(save_path):
        plt.savefig(save_path)
        print("Your graph was saved into {}".format(save_path))

    def errorbar_passband(
        self, t, flux, flux_err, passbands, passband, *, ax=None, title="", label="", marker="^", save=None
    ):
        if ax is None:
            ax = self._ax_adjust()

        anobject = self._make_dataframe(t, flux, flux_err, passbands)
        anobject = anobject.sort_values("time")
        light_curve = anobject[anobject["passband"] == passband]

        ax.errorbar(
            light_curve["time"].values,
            light_curve["flux"].values,
            yerr=light_curve["flux_err"].values,
            linewidth=3.5,
            fmt=marker,
            elinewidth=1.7,
            markersize=14.50,
            markeredgecolor="black",
            markeredgewidth=1.50,
            color=self.colors[passband],
            label="{} {}".format(passband, label),
        )

        ax.legend(loc="best", ncol=3, fontsize=20)
        ax.set_title(title, size=35, pad=15)

        if save is not None:
            self._save_fig(save)

        return ax

    def plot_sum_passbands(
        self, t_approx, flux_approx, flux_err_approx, passband_approx, *, ax=None, title="", save=None
    ):
        """ """

        if ax is None:
            ax = self._ax_adjust()
        if (
            (t_approx is not None)
            & (flux_approx is not None)
            & (flux_err_approx is not None)
            & (passband_approx is not None)
        ):
            anobject_approx = self._make_dataframe(t_approx, flux_approx, flux_err_approx, passband_approx)
            anobject_approx = anobject_approx.sort_values("time")
            curve = anobject_approx[["time", "flux"]].groupby("time", as_index=False).sum()
            ax.plot(curve["time"].values, curve["flux"].values, label="sum", linewidth=5.5, color="pink")
            ax.legend(loc="best", ncol=3, fontsize=20)
            ax.set_title(title, size=35, pad=15)
        if save is not None:
            self._save_fig(save)

        return ax

    def plot_peak(self, t_approx, flux_approx, flux_err_approx, passband_approx, *, ax=None, title="", save=None):
        """ """

        if ax is None:
            ax = self._ax_adjust()
        if (
            (t_approx is not None)
            & (flux_approx is not None)
            & (flux_err_approx is not None)
            & (passband_approx is not None)
        ):
            anobject_approx = self._make_dataframe(t_approx, flux_approx, flux_err_approx, passband_approx)
            curve = anobject_approx[["time", "flux"]].groupby("time", as_index=False).sum()
            pred_peak = curve["time"][curve["flux"].argmax()]
            ax.axvline(pred_peak, label="pred peak", color="red", linestyle="--", linewidth=5.5)
            ax.legend(loc="best", ncol=3, fontsize=20)
            ax.set_title(title, size=35, pad=15)
        if save is not None:
            self._save_fig(save)

        return ax

    def plot_approx(
        self, t_approx, flux_approx, flux_err_approx, passband_approx, *, passband, ax=None, title="", save=None
    ):
        """ """

        if ax is None:
            ax = self._ax_adjust()

        anobject_approx = self._make_dataframe(t_approx, flux_approx, flux_err_approx, passband_approx)
        anobject_approx = anobject_approx.sort_values("time")
        light_curve_approx = anobject_approx[anobject_approx.passband == passband]
        ax.plot(
            light_curve_approx["time"].values,
            light_curve_approx["flux"].values,
            linewidth=3.5,
            color=self.colors[passband],
            label="{} approx flux".format(passband),
            zorder=10,
        )
        ax.fill_between(
            light_curve_approx["time"].values,
            light_curve_approx["flux"].values - light_curve_approx["flux_err"].values,
            light_curve_approx["flux"].values + light_curve_approx["flux_err"].values,
            color=self.colors[passband],
            alpha=0.2,
            label="{} approx sigma".format(passband),
        )
        ax.legend(loc="best", ncol=3, fontsize=20)
        ax.set_title(title, size=35, pad=15)
        if save is not None:
            self._save_fig(save)

        return ax

    def plot_one_graph_passband(
        self,
        *,
        t,
        flux,
        flux_err,
        passbands,
        passband,
        ax=None,
        t_test=None,
        flux_test=None,
        flux_err_test=None,
        passband_test=None,
        t_approx=None,
        flux_approx=None,
        flux_err_approx=None,
        passband_approx=None,
        title="",
        save=None,
    ):
        """ """
        if ax is None:
            ax = self._ax_adjust()

        self.errorbar_passband(t=t, flux=flux, flux_err=flux_err, passbands=passbands, passband=passband, ax=ax)
        if (t_test is not None) & (flux_test is not None) & (flux_err_test is not None) & (passband_test is not None):
            self.errorbar_passband(t=t_test, flux=flux_test, flux_err=flux_err_test, passbands=passband_test)

        if (
            (t_approx is not None)
            & (flux_approx is not None)
            & (flux_err_approx is not None)
            & (passband_approx is not None)
        ):
            self.plot_approx(
                t_approx=t_approx,
                flux_approx=flux_approx,
                flux_err_approx=flux_err_approx,
                passband_approx=passband_approx,
                passband=passband,
                ax=ax,
            )

        ax.legend(loc="best", ncol=3, fontsize=20)
        ax.set_title(title, size=35, pad=15)
        if save is not None:
            self._save_fig(save)

        return ax

    def plot_true_peak(self, *, true_peak, ax=None, title="", save=None):
        """ """

        if ax is None:
            ax = self._ax_adjust()

        ax.axvline(true_peak, label="true peak", color="black", linewidth=5.5)
        ax.set_title(title, size=35, pad=15)
        if save is not None:
            self._save_fig(save)

        return ax

    def plot_one_graph_all(
        self,
        *,
        t,
        flux,
        flux_err,
        passbands,
        t_approx=None,
        flux_approx=None,
        flux_err_approx=None,
        passband_approx=None,
        passband=None,
        ax=None,
        true_peak=None,
        plot_peak=False,
        title="",
        save=None,
    ):
        """
        Plotting train points of light curve with errors for all passbands on one graph by default.

        If you submit the name passband, only the submitted passband is built.

        If you submit augmented data, a black solid curve is plotted at the predicted points.
        The predicted flux errors are also plotted using a gray bar.

        Parameters:
        -----------
        t_train, t_approx : array-like
            Timestamps of light curve observations, which are used in fit method\after augmentation.
        flux_train, flux_approx : array-like
            Flux of the light curve observations, which are used in fit method\after augmentation.
        flux_err_train, flux_err_approx : array-like
            Flux errors of the light curve observations, which are used in fit method\after augmentation.
        passband_train, passband_approx : array-like
            Passband IDs for each observation, which are used in fit method\after augmentation.
        passband : str or int or float
            Passband ID.
        ax : matplotlib.pyplot.subplot object
            You can set the axis as an element of your matplotlib.pyplot.figure object.
        true_peak : float or int
            The real peak of the light curve flux.
        plot_peak : bool or int
            Flag is responsible for plotting peak by max flux of overall flux.
        title : str
            The name of the graph set by ax.
        save : str
            The name for saving graph (in pdf format).
        """

        if ax is None:
            ax = self._ax_adjust()

        if passband is not None:
            self.plot_one_graph_passband(
                t=t,
                flux=flux,
                flux_err=flux_err,
                passbands=passbands,
                passband=passband,
                ax=ax,
                t_approx=t_approx,
                flux_approx=flux_approx,
                flux_err_approx=flux_err_approx,
                passband_approx=passband_approx,
            )

        else:
            for band in self.passband2lam.keys():
                self.plot_one_graph_passband(
                    t=t,
                    flux=flux,
                    flux_err=flux_err,
                    passbands=passbands,
                    passband=band,
                    ax=ax,
                    t_approx=t_approx,
                    flux_approx=flux_approx,
                    flux_err_approx=flux_err_approx,
                    passband_approx=passband_approx,
                )

        if true_peak is not None:
            self.plot_true_peak(true_peak, ax=ax)

        if plot_peak:
            self.plot_sum_passbands(
                t_approx=t_approx,
                flux_approx=flux_approx,
                flux_err_approx=flux_err_approx,
                passband_approx=passband_approx,
                ax=ax,
            )
            self.plot_peak(
                t_approx=t_approx,
                flux_approx=flux_approx,
                flux_err_approx=flux_err_approx,
                passband_approx=passband_approx,
                ax=ax,
            )

        ax.set_title(title, size=35, pad=15)
        ax.legend(loc="best", ncol=3, fontsize=20)
        if save is not None:
            self._save_fig(save)

        return ax



import numpy as np

from sklearn.neural_network import MLPRegressor
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import mean_squared_error

from fulu._base_aug import BaseAugmentation, add_log_lam


class MLPRegressionAugmentation(BaseAugmentation):
    """
    Light Curve Augmentation based on scikit-learn MLPRegressor

    Parameters:
    -----------
    passband2lam : dict
        A dictionary, where key is a passband ID and value is Log10 of its wave length.
        Example:
            passband2lam  = {0: np.log10(3751.36), 1: np.log10(4741.64), 2: np.log10(6173.23),
                             3: np.log10(7501.62), 4: np.log10(8679.19), 5: np.log10(9711.53)}
    hidden_layer_sizes : tuple
        The ith element represents the number of neurons in the ith hidden layer, length = n_layers - 2.
    solver : string
        The solver for weight optimization.
    activation : string
        Activation function for the hidden layer. Possible values: {‘identity’, ‘logistic’, ‘tanh’, ‘relu’}.
    learning_rate_init : float
        The initial learning rate used. It controls the step-size in updating the weights. Only used when solver=’sgd’
        or ‘adam’.
    max_iter : int
        Maximum number of iterations. The solver iterates until convergence or this number of iterations.
        For stochastic solvers (‘sgd’, ‘adam’), note that this determines the number of epochs
        (how many times each data point will be used), not the number of gradient steps.
    batch_size : int
        Size of minibatches for stochastic optimizers. If the solver is ‘lbfgs’,
        the classifier will not use minibatch. When set to “auto”, batch_size=min(200, n_samples)
    weight_decay : float
        L2 penalty (regularization term) parameter.
    """

    def __init__(
        self,
        passband2lam,
        hidden_layer_sizes=(
            20,
            10,
        ),
        solver="lbfgs",
        activation="tanh",
        learning_rate_init=0.001,
        max_iter=90,
        batch_size=1,
        weight_decay=0.0001,
    ):
        super().__init__(passband2lam)

        self.ss_x = None
        self.ss_y = None
        self.ss_t = None
        self.reg = None
        self.flux_err = 0

        self.hidden_layer_sizes = hidden_layer_sizes
        self.solver = solver
        self.activation = activation
        self.learning_rate_init = learning_rate_init
        self.max_iter = max_iter
        self.batch_size = batch_size
        self.weight_decay = weight_decay

    def _preproc_features(self, t, passband, ss_t):
        passband = np.array(passband)
        log_lam = add_log_lam(passband, self.passband2lam)
        t = ss_t.transform(np.array(t).reshape((-1, 1)))

        X = np.concatenate((t, log_lam.reshape((-1, 1))), axis=1)
        return X

    def fit(self, t, flux, flux_err, passband):
        """
        Fit an augmentation model.

        Parameters:
        -----------
        t : array-like
            Timestamps of light curve observations.
        flux : array-like
            Flux of the light curve observations.
        flux_err : array-like
            Flux errors of the light curve observations.
        passband : array-like
            Passband IDs for each observation.
        """

        super().fit(t, flux, flux_err, passband)

        self.ss_t = StandardScaler().fit(np.array(t).reshape((-1, 1)))

        X = self._preproc_features(t, passband, self.ss_t)
        self.ss_x = StandardScaler().fit(X)
        X_ss = self.ss_x.transform(X)
        flux = np.array(flux)

        self.ss_y = StandardScaler().fit(flux.reshape((-1, 1)))
        y_ss = self.ss_y.transform(flux.reshape((-1, 1)))

        self.reg = MLPRegressor(
            hidden_layer_sizes=self.hidden_layer_sizes,
            solver=self.solver,
            activation=self.activation,
            learning_rate_init=self.learning_rate_init,
            max_iter=self.max_iter,
            batch_size=self.batch_size,
            alpha=self.weight_decay,
        )
        self.reg.fit(X_ss, y_ss)

        flux_pred = self.ss_y.inverse_transform(self.reg.predict(X_ss).reshape(-1, 1)).reshape(-1)
        self.flux_err = np.sqrt(mean_squared_error(flux, flux_pred))
        return self

    def predict(self, t, passband):
        """
        Apply the augmentation model to the given observation time moments.

        Parameters:
        -----------
        t : array-like
            Timestamps of light curve observations.
        passband : array-like
            Passband IDs for each observation.

        Returns:
        --------
        flux_pred : array-like
            Flux of the light curve observations, approximated by the augmentation model.
        flux_err_pred : array-like
            Flux errors of the light curve observations, estimated by the augmentation model.
        """

        X = self._preproc_features(t, passband, self.ss_t)
        X_ss = self.ss_x.transform(X)

        flux_pred = self.ss_y.inverse_transform(self.reg.predict(X_ss).reshape(-1, 1)).reshape(-1)
        flux_err_pred = np.full_like(flux_pred, self.flux_err)

        return np.maximum(flux_pred, np.zeros(flux_pred.shape)), flux_err_pred



import numpy as np
from sklearn.preprocessing import StandardScaler
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import TensorDataset, DataLoader
from torch.autograd import Variable

from fulu._base_aug import BaseAugmentation, add_log_lam


# TODO: consider to remove
class InvertibleLayer(nn.Module):
    def __init__(self, var_size):
        super().__init__()

        self.var_size = var_size


class NormalizingFlow(nn.Module):
    def __init__(self, layers, prior):
        super().__init__()

        self.layers = nn.ModuleList(layers)
        self.prior = prior

    def log_prob(self, x, y):
        """
        x: torch.Tensor of shape [batch_size, var_size]
            Data
        y: torch.Tensor of shape [batch_size, cond_size]
            Condition
        """
        log_likelihood = None

        for layer in self.layers:
            x, change = layer.f(x, y)
            if log_likelihood is not None:
                log_likelihood = log_likelihood + change
            else:
                log_likelihood = change
        log_likelihood = log_likelihood + self.prior.log_prob(x)

        return log_likelihood.mean()

    def sample(self, y):
        """
        y: torch.Tensor of shape [batch_size, cond_size]
            Condition
        """

        x = self.prior.sample((len(y),))
        for layer in self.layers[::-1]:
            x = layer.g(x, y)

        return x


class RealNVP(InvertibleLayer):
    def __init__(self, var_size, cond_size, mask, hidden=10):
        super().__init__(var_size=var_size)

        self.mask = mask

        self.nn_t = nn.Sequential(nn.Linear(var_size + cond_size, hidden), nn.Tanh(), nn.Linear(hidden, var_size))
        self.nn_s = nn.Sequential(
            nn.Linear(var_size + cond_size, hidden),
            nn.Tanh(),
            nn.Linear(hidden, var_size),
        )

    def f(self, x, y):
        """
        x: torch.Tensor of shape [batch_size, var_size]
            Data
        y: torch.Tensor of shape [batch_size, cond_size]
            Condition
        """
        xy = torch.cat((x * self.mask[None, :], y), dim=1)
        t = self.nn_t(xy)
        s = self.nn_s(xy)

        new_x = (x * torch.exp(s) + t) * (1 - self.mask[None, :]) + x * self.mask[None, :]
        log_det = (s * (1 - self.mask[None, :])).sum(dim=-1)
        return new_x, log_det

    def g(self, x, y):
        """
        x: torch.Tensor of shape [batch_size, var_size]
            Data
        y: torch.Tensor of shape [batch_size, cond_size]
            Condition
        """
        xy = torch.cat((x * self.mask[None, :], y), dim=1)
        t = self.nn_t(xy)
        s = self.nn_s(xy)

        new_x = ((x - t) * torch.exp(-s)) * (1 - self.mask[None, :]) + x * self.mask[None, :]
        return new_x


class NFFitter(object):
    def __init__(
        self,
        var_size=2,
        cond_size=2,
        normalize_y=True,
        batch_size=32,
        n_epochs=10,
        lr=0.0001,
        randomize_x=True,
        device="cpu",
        weight_decay=0,
    ):

        self.normalize_y = normalize_y
        self.randomize_x = randomize_x
        self.batch_size = batch_size
        self.n_epochs = n_epochs
        self.lr = lr
        self.weight_decay = weight_decay

        prior = torch.distributions.MultivariateNormal(torch.zeros(var_size), torch.eye(var_size))

        layers = []
        for i in range(8):
            layers.append(
                RealNVP(var_size=var_size, cond_size=cond_size + randomize_x, mask=((torch.arange(var_size) + i) % 2))
            )

        self.nf = NormalizingFlow(layers=layers, prior=prior)
        self.opt = torch.optim.Adam(self.nf.parameters(), lr=self.lr, weight_decay=self.weight_decay)

        self.device = torch.device(device)

    def reshape(self, y):
        if y.ndim < 2:
            return y.reshape(-1, 1)
        return y

    def fit(self, X, y, y_std=None):

        # reshape
        y = self.reshape(y)

        # normalize
        if self.normalize_y:
            self.ss_y = StandardScaler()
            y = self.ss_y.fit_transform(y)

        if y_std is not None:
            y_std = self.reshape(y_std)
            if self.normalize_y:
                y_std /= self.ss_y.scale_
        else:
            y_std = np.zeros_like(y)

        # numpy to tensor
        y_real = torch.tensor(y, dtype=torch.float32, device=self.device)
        y_real_std = torch.tensor(y_std, dtype=torch.float32, device=self.device)
        X_cond = torch.tensor(X, dtype=torch.float32, device=self.device)

        # tensor to dataset
        dataset_real = TensorDataset(y_real, y_real_std, X_cond)

        # Fit GAN
        for epoch in range(self.n_epochs):
            for i, (y_batch, std_batch, x_batch) in enumerate(
                DataLoader(dataset_real, batch_size=self.batch_size, shuffle=True)
            ):
                noise = np.random.normal(0, 1, (len(y_batch), 1))
                noise = torch.tensor(noise, dtype=torch.float32, device=self.device)

                y_batch = torch.normal(y_batch, std_batch)
                y_batch = torch.cat((y_batch, noise), dim=1)

                if self.randomize_x:
                    noise = np.random.normal(0, 1, (len(x_batch), 1))
                    noise = torch.tensor(noise, dtype=torch.float32, device=self.device)
                    x_batch = torch.cat((x_batch, noise), dim=1)

                # caiculate loss
                loss = -self.nf.log_prob(y_batch, x_batch)

                # optimization step
                self.opt.zero_grad()
                loss.backward()
                self.opt.step()

    def predict(self, X):
        X = torch.tensor(X, dtype=torch.float32, device=self.device)
        if self.randomize_x:
            noise = np.random.normal(0, 1, (len(X), 1))
            noise = torch.tensor(noise, dtype=torch.float32, device=self.device)
            X = torch.cat((X, noise), dim=1)
        y_pred = self.nf.sample(X).cpu().detach().numpy()[:, 0]
        # normalize
        if self.normalize_y:
            y_pred = self.ss_y.inverse_transform(y_pred.reshape(-1, 1)).reshape(-1)
        return y_pred

    def predict_n_times(self, X, n_times=100):
        predictions = []
        for i in range(n_times):
            y_pred = self.predict(X).reshape(-1)
            predictions.append(y_pred)
        predictions = np.array(predictions)
        mean = predictions.mean(axis=0)
        std = predictions.std(axis=0)
        return mean, std


class NormalizingFlowAugmentation(BaseAugmentation):
    """
    Light Curve Augmentation based on Normalizing Flows Regression

    Parameters:
    -----------
    passband2lam : dict
        A dictionary, where key is a passband ID and value is Log10 of its wave length.
        Example:
            passband2lam  = {0: np.log10(3751.36), 1: np.log10(4741.64), 2: np.log10(6173.23),
                             3: np.log10(7501.62), 4: np.log10(8679.19), 5: np.log10(9711.53)}
    batch_size : int
        A number of observations to take for the model optimization.
        Example:
            batch_size = 500
    n_epochs : int
        A number of epoches for the model optimization.
        Example:
            n_epochs = 3000
    lr : float
        A learning rate value of the optimization.
        Example:
            lr = 0.005
    weight_decay : float
        L2 penalty (regularization term) parameter.
    """

    def __init__(self, passband2lam, batch_size=500, n_epochs=3000, lr=0.005, device="cpu", weight_decay=0):
        super().__init__(passband2lam)

        self.batch_size = batch_size
        self.n_epochs = n_epochs
        self.lr = lr
        self.device = device
        self.weight_decay = weight_decay

        self.passband2lam = passband2lam

        self.ss = None
        self.reg = None

    def fit(self, t, flux, flux_err, passband):
        """
        Fit an augmentation model.

        Parameters:
        -----------
        t : array-like
            Timestamps of light curve observations.
        flux : array-like
            Flux of the light curve observations.
        flux_err : array-like
            Flux errors of the light curve observations.
        passband : array-like
            Passband IDs for each observation.
        """

        super().fit(t, flux, flux_err, passband)

        t = np.array(t)
        flux = np.array(flux)
        flux_err = np.array(flux_err)
        passband = np.array(passband)
        log_lam = add_log_lam(passband, self.passband2lam)

        X = np.concatenate((t.reshape(-1, 1), log_lam.reshape(-1, 1)), axis=1)

        self.ss = StandardScaler()
        X_ss = self.ss.fit_transform(X)

        self.reg = NFFitter(
            var_size=2,
            cond_size=2,
            normalize_y=True,
            batch_size=self.batch_size,
            n_epochs=self.n_epochs,
            lr=self.lr,
            randomize_x=True,
            device=self.device,
            weight_decay=self.weight_decay,
        )
        self.reg.fit(X_ss, flux, flux_err)

        return self

    def predict(self, t, passband):
        """
        Apply the augmentation model to the given observation time moments.

        Parameters:
        -----------
        t : array-like
            Timestamps of light curve observations.
        passband : array-like
            Passband IDs for each observation.

        Returns:
        --------
        flux_pred : array-like
            Flux of the light curve observations, approximated by the augmentation model.
        flux_err_pred : array-like
            Flux errors of the light curve observations, estimated by the augmentation model.
        """

        t = np.array(t)
        passband = np.array(passband)
        log_lam = add_log_lam(passband, self.passband2lam)

        X = np.concatenate((t.reshape(-1, 1), log_lam.reshape(-1, 1)), axis=1)
        X_ss = self.ss.transform(X)

        flux_pred, flux_err_pred = self.reg.predict_n_times(X_ss, n_times=100)

        return flux_pred, flux_err_pred



import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchbnn as bnn
from torch.utils.data import TensorDataset, DataLoader
from sklearn.preprocessing import StandardScaler

from fulu._base_aug import BaseAugmentation, add_log_lam


class BNNRegressor(nn.Module):
    def __init__(self, n_inputs=1, n_hidden=10, prior_sigma=0.01):
        super(BNNRegressor, self).__init__()

        self.model = nn.Sequential(
            bnn.BayesLinear(prior_mu=0, prior_sigma=prior_sigma, in_features=n_inputs, out_features=n_hidden),
            nn.Tanh(),
            # bnn.BayesLinear(prior_mu=0, prior_sigma=prior_sigma, in_features=n_hidden, out_features=n_hidden),
            # nn.Tanh(),
            bnn.BayesLinear(prior_mu=0, prior_sigma=prior_sigma, in_features=n_hidden, out_features=1),
        )

    def forward(self, x):
        return self.model(x)


class FitBNNRegressor:
    def __init__(
        self,
        n_hidden=10,
        prior_sigma=0.01,
        n_epochs=10,
        lr=0.01,
        kl_weight=0.1,
        optimizer="Adam",
        debug=0,
        device="cpu",
        weight_decay=0,
    ):
        self.model = None
        self.n_hidden = n_hidden
        self.prior_sigma = prior_sigma
        self.n_epochs = n_epochs
        self.lr = lr
        self.kl_weight = kl_weight
        self.optimizer = optimizer
        self.debug = debug
        self.device = torch.device(device)
        self.weight_decay = weight_decay

    def fit(self, X, y):
        # Estimate model
        self.model = BNNRegressor(n_inputs=X.shape[1], n_hidden=self.n_hidden, prior_sigma=self.prior_sigma)
        self.model.to(self.device)
        # Convert X and y into torch tensors
        X_tensor = torch.as_tensor(X, dtype=torch.float32, device=self.device)
        y_tensor = torch.as_tensor(y.reshape(-1, 1), dtype=torch.float32, device=self.device)
        # Estimate loss
        mse_loss = nn.MSELoss()
        kl_loss = bnn.BKLLoss(reduction="mean", last_layer_only=False)
        # Estimate optimizer
        if self.optimizer == "Adam":
            opt = torch.optim.Adam(self.model.parameters(), lr=self.lr, weight_decay=self.weight_decay)
        elif self.optimizer == "SGD":
            opt = torch.optim.SGD(self.model.parameters(), lr=self.lr, weight_decay=self.weight_decay)
        elif self.optimizer == "RMSprop":
            opt = torch.optim.RMSprop(self.model.parameters(), lr=self.lr, weight_decay=self.weight_decay)
        else:
            opt = torch.optim.Adam(self.model.parameters(), lr=self.lr, weight_decay=self.weight_decay)
        # Enable droout
        self.model.train(True)
        # Start the model fit
        for epoch_i in range(self.n_epochs):
            y_pred = self.model(X_tensor)
            mse = mse_loss(y_pred, y_tensor)
            kl = kl_loss(self.model)
            loss = mse + self.kl_weight * kl
            opt.zero_grad()
            loss.backward()
            opt.step()

    def predict(self, X, n_times=1):
        # Disable droout
        self.model.train(False)
        # Convert X and y into torch tensors
        X_tensor = torch.as_tensor(X, dtype=torch.float32, device=self.device)
        # Make predictions for X
        y_pred = self.model(X_tensor)
        y_pred = y_pred.cpu().detach().numpy()
        return y_pred

    def predict_n_times(self, X, ss, n_times=100):
        predictions = []
        for i in range(n_times):
            y_pred = self.predict(X).reshape(
                -1,
            )
            predictions.append(y_pred)
        predictions = ss.inverse_transform(np.array(predictions))
        mean = predictions.mean(axis=0)
        std = predictions.std(axis=0)
        return mean, std


class BayesianNetAugmentation(BaseAugmentation):
    """
    Light Curve Augmentation based on BNNRegressor with new features

    Parameters:
    -----------
    passband2lam : dict
        A dictionary, where key is a passband ID and value is Log10 of its wave length.
        Example:
            passband2lam  = {0: np.log10(3751.36), 1: np.log10(4741.64), 2: np.log10(6173.23),
                             3: np.log10(7501.62), 4: np.log10(8679.19), 5: np.log10(9711.53)}
    n_hidden : int
        A number of hidden neurons in the network.
    prior_sigma : float
        The standard deviation for a neuron weights prior distribution.
    n_epochs : int
        The number of epochs in training.
    lr : float
        The learning rate for the network optimization.
    kl_weight : float
        The coefficient for a regularization term of a loss function.
    optimizer : string
        Optimization algorithm name. Possible values: 'Adam', 'RMSprop', 'SGD'.
    device : str or torch device
        Torch device name, default is 'cpu'
    weight_decay : float
        L2 penalty (regularization term) parameter.
    """

    def __init__(
        self,
        passband2lam,
        n_hidden=20,
        prior_sigma=0.1,
        n_epochs=3000,
        lr=0.01,
        kl_weight=0.0001,
        optimizer="Adam",
        device="cpu",
        weight_decay=0,
    ):
        super().__init__(passband2lam)

        self.n_hidden = n_hidden
        self.prior_sigma = prior_sigma
        self.n_epochs = n_epochs
        self.lr = lr
        self.kl_weight = kl_weight
        self.optimizer = optimizer
        self.device = device
        self.weight_decay = weight_decay

        self.ss_x = None
        self.ss_y = None
        self.reg = None

    def _preproc_features(self, t, passband):
        t = np.array(t).reshape((-1, 1))
        passband = np.array(passband)
        log_lam = add_log_lam(passband, self.passband2lam).reshape((-1, 1))

        X = np.concatenate((t, log_lam), axis=1)
        return X

    def fit(self, t, flux, flux_err, passband):
        """
        Fit an augmentation model.

        Parameters:
        -----------
        t : array-like
            Scaled timestamps of light curve observations.
        flux : array-like
            Flux of the light curve observations.
        flux_err : array-like
            Flux errors of the light curve observations.
        passband : array-like
            Passband IDs for each observation.
        """
        super().fit(t, flux, flux_err, passband)

        X = self._preproc_features(t, passband)
        self.ss_x = StandardScaler().fit(X)
        X_ss = self.ss_x.transform(X)
        flux = np.array(flux)

        self.ss_y = StandardScaler().fit(flux.reshape((-1, 1)))
        y_ss = self.ss_y.transform(flux.reshape((-1, 1)))
        self.reg = FitBNNRegressor(
            n_hidden=self.n_hidden,
            prior_sigma=self.prior_sigma,
            n_epochs=self.n_epochs,
            lr=self.lr,
            kl_weight=self.kl_weight,
            optimizer=self.optimizer,
            device=self.device,
            weight_decay=self.weight_decay,
        )
        self.reg.fit(X_ss, y_ss)
        return self

    def predict(self, t, passband):
        """
        Apply the augmentation model to the given observation time moments.

        Parameters:
        -----------
        t : array-like
            Scaled timestamps of light curve observations.
        passband : array-like
            Passband IDs for each observation.

        Returns:
        --------
        flux_pred : array-like
            Flux of the light curve observations, approximated by the augmentation model.
        flux_err_pred : array-like
            Flux errors of the light curve observations, estimated by the augmentation model.
        """

        X = self._preproc_features(t, passband)
        X_ss = self.ss_x.transform(X)

        flux_pred, flux_err_pred = self.reg.predict_n_times(X_ss, self.ss_y)

        return np.maximum(np.zeros(flux_pred.shape), flux_pred), flux_err_pred



from .bnn_aug import BayesianNetAugmentation
from .gp_aug import GaussianProcessesAugmentation
from .mlp_reg_aug import MLPRegressionAugmentation
from .single_layer_aug import SingleLayerNetAugmentation
from .nf_aug import NormalizingFlowAugmentation

from .plotting import LcPlotter
